/**
 * @file outside.js
 * @desc click指令
 * @author
 */
import Vue from 'vue';

/**
 * 判断是否是对象类型
 * @param {Object} value 值
 * @return {boolean} 是否
 */
function isObject(value) {
    return value != null && (typeof value === 'object' || typeof value === 'function');
}

/**
 * 判断是否是字符串类型
 * @param {Object} str 值
 * @return {boolean} 是否
 */
function isString(str) {
    return typeof str === 'string' && str.constructor === String;
}

/**
 * 判断是否是字符串类型vnode
 * @param {Object} vnode 值
 * @return {boolean} 是否
 */
function isVnode(vnode) {
    return isObject(vnode) && vnode.componentOptions;
}

/**
 * 查找node元素
 * @param {string} ref ref值
 * @param {Object} context 上下文
 * @return {Object} 元素
 */
function getVnodes(ref, context) {
    if (!ref) {
        return [];
    }
    let vnodes;
    if (isString(ref)) {
        vnodes = context.$refs[ref];
        if (!vnodes) {
            vnodes = [];
        } else {
            vnodes = Array.isArray(vnodes) ? vnodes : [vnodes];
        }
        vnodes = vnodes.map(item => item.$vnode || item);
    } else {
        ref = Array.isArray(ref) ? ref : [ref];
        vnodes = ref.map(item => {
            let node = null;
            if (item.$vnode) {
                node = item.$vnode;
            } else if (isVnode(item) || item.nodeType === 1 || item.nodeType === 3) {
                // vnode节点、dom元素节点和文本节点
                node = item;
            }
            return node;
        });
    }
    return vnodes || [];
}

/**
 * 根据ref拿到指定的DOM节点，主要有一下情况：
 * 1、如果ref是字符串，就需要从context下面找指定ref的组件实例或者DOM元素；
 * 2、如果ref直接就是组件实例，那就取ref.$el；
 * 3、如果ref直接就是文本节点或者元素节点，就返回ref。
 * 4、否则返回空数组
 *
 * @param {string|Vue|Node|Array<string>|Array<Vue>|Array<Node>} ref 目标节点标识
 * @param {VueContext=} context 组件上下文，在ref为字符串的时候必传
 * @return {Node}
 */
function getNodes(ref, context) {
    let vnodes = getVnodes(ref, context);
    return vnodes.map(item => {
        if (isVnode(item)) {
            return item.elm;
        }
        return item;
    });
}

/**
 * 通过ref寻找元素
 * @param {Array} refs ref数组
 * @param {Object} context 上下文
 * @return {Object} 元素
 */
function getElementsByRefs(refs, context) {
    const elements = [];
    refs.forEach(ref => {
        elements.push(...getNodes(ref, context));
    });
    return elements;
}

/**
 * 判断两个元素是否存在父子关系。
 * IE9 的 SVGSVGElement 上没有 contains 方法，做下 hack 。
 *
 * @param {Element} parentElem 父元素
 * @param {Element} childElem 子元素
 * @return {boolean}
 */
function contains(parentElem, childElem) {
    return parentElem.contains
      ? parentElem.contains(childElem)
      : document.body.contains.call(parentElem, childElem);
}

/**
 * 判断 element 在 DOM 树结构上是否被包含在 elements 里面
 *
 * @param {Element} element 待判断的元素
 * @param {Array.<Element>} elements 元素范围
 * @return {boolean}
 */
function isElementIn(element, elements) {
    return elements.some(elm => contains(elm, element));
}

/**
 * 绑定指令函数
 * @param {Object} el 指令绑定的元素
 * @param {Object} binding 绑定值属性的对象
 * @param {Object} vnode node节点
 */
function load(el, binding, vnode) {
    let {value, oldValue, arg} = binding;
    let refs = arg ? arg.split(',') : [];
    // 如果当前值与前一值相等，则不再绑定
    if (value === oldValue) {
        return;
    }
    if (typeof value !== 'function') {
        throw new Error('the click events must be function');
    } else if (el.addEventListener) {
        oldValue && document.body.removeEventListener('click', oldValue);
        document.body.addEventListener('click', e => {
            if (!isElementIn(e.target, getElementsByRefs(refs, vnode.context))) {
                value && value(e);
            }
        });
    } else if (document.body.attachEvent) {
        oldValue && document.body.detachEvent('onclick', oldValue);
        document.body.attachEvent('onclick', e => {
            if (!isElementIn(e.target, getElementsByRefs(refs, vnode.context))) {
                value && value(e);
            }
        });
    }
}

/**
 * 解绑
 * @param {Object} el 指令绑定的元素
 * @param {Object} value 指令绑定的值
 */
function clear(el, {value}) {
    if (value && typeof value === 'function') {
        if (document.body.removeEventListener) {
            document.body.removeEventListener('click', value);
        } else {
            document.body.detachEvent('onclick', value);
        }
    }
}

// 导出scroll指令
export default {
    inserted(el, binding, vnode) {
        load(el, binding, vnode);
    },
    componentUpdated(el, binding, vnode) {
        load(el, binding, vnode);
    },
    unbind(el, binding) {
        clear(el, binding);
    }
};
